from __future__ import annotations

from abc import abstractmethod
from typing import Any, AsyncContextManager, Awaitable, Generic, Union

from patio.registry import Registry, T, TaskFunctionType


class AbstractExecutor(Generic[T], AsyncContextManager):
    """
    An Executor is an entity that executes local functions on the local side.
    The following executors are implemented in the package:

    * :class:`AsyncExecutor`
    * :class:`ThreadPoolExecutor`
    * :class:`ProcessPoolExecutor`
    * :class:`NullExecutor`

    Its role is to reliably execute jobs without taking too much so as not to
    cause a denial of service, or excessive memory consumption.

    The executor instance is passing to the broker, it's usually applies
    it to the whole registry. Therefore, you should understand what functions
    the registry must contain to choose kind of an executor.
    """

    DEFAULT_MAX_WORKERS: int = 4

    def __init__(
        self, registry: Registry, max_workers: int = DEFAULT_MAX_WORKERS,
    ):
        self.registry = registry
        self.max_workers = max_workers

    @abstractmethod
    async def setup(self) -> None:
        """
        Configures the executor, can be called several times, with the
        assumption that it will be configured exactly once.
        """
        raise NotImplementedError(
            f"Not implemented method setup "
            f"in {self.__class__.__name__!r} ",
        )

    @abstractmethod
    def submit(
        self, func: TaskFunctionType, *args: Any, **kwargs: Any
    ) -> Awaitable[T]:
        """
        Passes the function to execute, and waits for the result, returns it.

        :param func: Function to be executed
        :param args: positional arguments of the Function
        :param kwargs: keyword arguments of the function
        :return: the result that the function will return
        """
        raise NotImplementedError(
            f"Not implemented method submit in {self.__class__.__name__!r} "
            f"Call {func!r} with args={args!r}, kwargs={kwargs!r} skipped",
        )

    @abstractmethod
    async def shutdown(self) -> None:
        """
        Performs an executor stop. All related and unperformed tasks
        must be completed.
        """

    async def __aenter__(self) -> AbstractExecutor:
        await self.setup()
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.shutdown()

    async def execute(
        self, func: Union[str, TaskFunctionType], *args: Any, **kwargs: Any
    ) -> T:
        func = self.registry.resolve(func)
        return await self.submit(func, *args, ** kwargs)



from patio.broker import (
    AbstractBroker, MemoryBroker, TCPBrokerBase, TCPClientBroker,
    TCPServerBroker,
)
from patio.executor import (
    AbstractExecutor, AsyncExecutor, NullExecutor, ProcessPoolExecutor,
    ThreadPoolExecutor,
)
from patio.registry import Registry, TaskFunctionType


__all__ = (
    "AbstractBroker",
    "AbstractExecutor",
    "AsyncExecutor",
    "MemoryBroker",
    "NullExecutor",
    "ProcessPoolExecutor",
    "Registry",
    "TCPBrokerBase",
    "TCPClientBroker",
    "TCPServerBroker",
    "TaskFunctionType",
    "ThreadPoolExecutor",
)



import asyncio
import concurrent.futures
from functools import cached_property, partial
from typing import Any, Awaitable

from patio.executor.base import AbstractExecutor
from patio.registry import SyncTaskFunctionType, T


class ProcessPoolExecutor(AbstractExecutor[T]):
    """
    Execute jobs in the process pool. Jobs cannot be asynchronous functions.
    This means that the whole registry must not contain functions other than
    specified kind.
    """
    __slots__ = "max_workers", "executor"

    DEFAULT_MAX_WORKERS: int = 4

    executor: concurrent.futures.ProcessPoolExecutor

    @cached_property
    def loop(self) -> asyncio.AbstractEventLoop:
        return asyncio.get_running_loop()

    async def setup(self) -> None:
        if hasattr(self, "executor"):
            return

        self.executor = await self.loop.run_in_executor(
            None, concurrent.futures.ProcessPoolExecutor, self.max_workers,
        )

    def submit(
        self, func: SyncTaskFunctionType, *args: Any, **kwargs: Any
    ) -> Awaitable[T]:
        return self.loop.run_in_executor(
            self.executor, partial(func, *args, **kwargs),
        )

    async def shutdown(self) -> None:
        await self.loop.run_in_executor(None, self.executor.shutdown, True)



from __future__ import annotations

import asyncio
import logging
import pickle
import ssl
import uuid
from abc import ABC, abstractmethod
from collections import deque
from dataclasses import dataclass
from functools import cached_property
from types import MappingProxyType
from typing import (
    Any, Callable, Coroutine, Deque, Dict, Mapping, Optional, Set, Tuple,
    TypeVar, Union, final,
)

from patio.broker import AbstractBroker, TimeoutType
from patio.broker.serializer import (
    AbstractSerializer, RestrictedPickleSerializer,
)
from patio.broker.tcp.protocol import Header, PacketTypes, Protocol
from patio.compat import Queue
from patio.executor import AbstractExecutor
from patio.registry import Registry, TaskFunctionType


T = TypeVar("T")
log = logging.getLogger(__name__)


@dataclass(frozen=True)
class RPCEvent:
    header: Header
    payload: bytes


@dataclass
class CallRequest:
    func: str
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]
    timeout: Optional[TimeoutType]


class PacketHandler(ABC):
    reader: asyncio.StreamReader
    writer: asyncio.StreamWriter
    protocol: Protocol
    executor: AbstractExecutor
    registry: Registry

    def __init__(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        executor: AbstractExecutor,
        protocol: Protocol,
    ):
        self.reader = reader
        self.writer = writer
        self.protocol = protocol
        self.executor = executor
        self.__client_info = self.writer.get_extra_info("peername")

        self.__events: Queue[RPCEvent] = Queue(self.executor.max_workers)
        self.__results: Dict[int, asyncio.Future] = {}

        self.__method_map: Mapping[
            PacketTypes, Callable[[RPCEvent], Coroutine[Any, Any, Any]],
        ] = MappingProxyType({
            PacketTypes.REQUEST: self.handle_request,
            PacketTypes.RESPONSE: self.handle_response,
            PacketTypes.ERROR: self.handle_error,
        })

    @cached_property
    def loop(self) -> asyncio.AbstractEventLoop:
        return asyncio.get_running_loop()

    async def get_event(self) -> RPCEvent:
        header = Header.unpack(await self.reader.readexactly(Header.SIZE))
        payload = await self.reader.readexactly(header.size)
        return RPCEvent(header=header, payload=payload)

    async def handle_request(self, event: RPCEvent) -> None:
        try:
            request: CallRequest = self.protocol.serializer.unpack(
                event.payload,
            )
            result = await asyncio.wait_for(
                self.executor.execute(
                    request.func, *request.args, **request.kwargs
                ),
                timeout=request.timeout,
            )
            self.writer.write(
                self.protocol.pack(
                    result, PacketTypes.RESPONSE, serial=event.header.serial,
                ),
            )
        except Exception as e:
            self.writer.write(
                self.protocol.pack(
                    e, PacketTypes.ERROR, serial=event.header.serial,
                ),
            )

    async def handle_response(self, event: RPCEvent) -> None:
        future = self.__results.pop(event.header.serial)
        if future.done():
            return
        try:
            payload = self.protocol.serializer.unpack(event.payload)
        except (ValueError, pickle.UnpicklingError) as e:
            future.set_exception(e)
            return

        future.set_result(payload)

    async def handle_error(self, event: RPCEvent) -> None:
        future = self.__results.pop(event.header.serial)
        if future.done():
            return

        try:
            payload = self.protocol.serializer.unpack(event.payload)
        except (ValueError, pickle.UnpicklingError) as e:
            future.set_exception(e)
            return

        future.set_exception(payload)

    async def step(self) -> None:
        event = await self.get_event()
        method = self.__method_map[event.header.type]
        await method(event)

    async def make_request(self, request: CallRequest) -> Any:
        serial = self.protocol.get_serial()
        future = self.loop.create_future()
        self.__results[serial] = future
        self.writer.write(
            self.protocol.pack(request, PacketTypes.REQUEST, serial=serial),
        )
        return await future

    @abstractmethod
    async def authorize(self) -> bool:
        raise NotImplementedError

    async def close(self) -> None:
        if not self.writer.can_write_eof():
            return
        self.writer.close()
        await self.writer.wait_closed()

    async def process_events(self) -> None:
        while True:
            event = await self.__events.get()
            method = self.__method_map[event.header.type]
            await method(event)
            self.__events.task_done()

    async def start_processing(self) -> None:
        workers = []

        for _ in range(self.executor.max_workers):
            workers.append(self.loop.create_task(self.process_events()))

        try:
            while True:
                try:
                    await self.__events.put(await self.get_event())
                except (ConnectionError, asyncio.IncompleteReadError):
                    address, port = self.__client_info[:2]

                    if ":" in address:
                        address = f"[{address}]"

                    log.info(
                        "Client connection tcp://%s:%d had been closed",
                        address, port,
                    )
                    return
        finally:
            del self.__events

            cancelling = []
            for worker in workers:
                if worker.done():
                    continue
                worker.cancel()
                cancelling.append(worker)
            if cancelling:
                await asyncio.gather(*cancelling, return_exceptions=True)


class ServerPacketHandler(PacketHandler):
    async def authorize(self) -> bool:
        token = uuid.uuid4().bytes
        self.writer.write(self.protocol.pack(token, PacketTypes.AUTH_DIGEST))
        event = await self.get_event()

        if event.header.type != PacketTypes.AUTH_REQUEST:
            return False

        if not self.protocol.authorize_check(event.payload):
            return False

        self.writer.write(self.protocol.pack(None, PacketTypes.AUTH_OK))
        return True


class ClientPacketHandler(PacketHandler):
    async def authorize(self) -> bool:
        event = await self.get_event()
        if event.header.type != PacketTypes.AUTH_DIGEST:
            return False
        token = event.payload
        self.writer.write(self.protocol.authorize_request(token))
        event = await self.get_event()
        return event.header.type == PacketTypes.AUTH_OK


class TCPBrokerBase(AbstractBroker, ABC):
    protocol: Protocol
    registry: Registry

    def __init__(
        self,
        executor: AbstractExecutor,
        key: bytes = b"", *,
        ssl_context: Optional[ssl.SSLContext] = None,
        reconnect_timeout: TimeoutType = 1,
        serializer: AbstractSerializer = RestrictedPickleSerializer()
    ):
        self.protocol = Protocol(key=key, serializer=serializer)
        self.reconnect_timeout = reconnect_timeout
        self._ssl_context: Optional[ssl.SSLContext] = ssl_context
        self.__handlers: Deque[PacketHandler] = deque()
        self.__rotate_lock = asyncio.Lock()

        super().__init__(executor=executor)

    async def _get_handler(self) -> PacketHandler:
        handler: PacketHandler

        async with self.__rotate_lock:
            while not self.__handlers:
                log.warning(
                    "No active connections, retrying after %.3f seconds.",
                    self.reconnect_timeout,
                )
                await asyncio.sleep(self.reconnect_timeout)

            handler = self.__handlers.popleft()
            self.__handlers.append(handler)
        return handler

    async def _add_handler(self, handler: PacketHandler) -> None:
        self.__handlers.append(handler)

    async def _remove_handler(self, handler: PacketHandler) -> None:
        async with self.__rotate_lock:
            self.__handlers.remove(handler)

    async def call(
        self,
        func: Union[str, TaskFunctionType],
        *args: Any,
        timeout: Optional[TimeoutType] = None,
        **kwargs: Any,
    ) -> Any:
        if not isinstance(func, str):
            raise TypeError("Only strings supports")

        request = CallRequest(
            func=func, args=args, kwargs=kwargs, timeout=timeout,
        )

        async def go() -> Any:
            handler = await self._get_handler()
            return await handler.make_request(request)

        return await asyncio.wait_for(go(), timeout=timeout)


DEFAULT_PORT = 15383


@final
class TCPServerBroker(TCPBrokerBase):
    def __init__(
        self,
        executor: AbstractExecutor,
        key: bytes = b"", *,
        ssl_context: Optional[ssl.SSLContext] = None,
        reconnect_timeout: TimeoutType = 1,
        serializer: AbstractSerializer = RestrictedPickleSerializer()
    ):
        super().__init__(
            executor=executor, key=key, ssl_context=ssl_context,
            reconnect_timeout=reconnect_timeout, serializer=serializer,
        )
        self.servers: Set[asyncio.AbstractServer] = set()

    async def _on_client_connected(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter,
    ) -> None:
        handler = ServerPacketHandler(
            reader=reader, writer=writer,
            executor=self.executor, protocol=self.protocol,
        )

        async def handle() -> None:
            if not await handler.authorize():
                raise RuntimeError("Unauthorized")

            await self._add_handler(handler)

            try:
                await handler.start_processing()
            finally:
                await self._remove_handler(handler)

        await self.create_task(handle())

    async def listen(
        self, address: str, port: int = DEFAULT_PORT, **kwargs: Any
    ) -> None:
        if "ssl" not in kwargs:
            kwargs["ssl"] = self._ssl_context

        self.servers.add(
            await asyncio.start_server(
                self._on_client_connected,
                host=address,
                port=port,
                **kwargs
            ),
        )

    async def close(self) -> None:
        tasks = [super().close()]
        for server in self.servers:
            server.close()
            tasks.append(server.wait_closed())
        await asyncio.gather(*tasks, return_exceptions=True)


@final
class TCPClientBroker(TCPBrokerBase):
    async def connection_fabric(
        self, address: str, port: int,
        on_connected: asyncio.Event,
        **kwargs: Any
    ) -> None:
        if "ssl" not in kwargs:
            kwargs["ssl"] = self._ssl_context

        while True:
            try:
                reader, writer = await asyncio.open_connection(
                    host=address, port=port, **kwargs
                )
                try:
                    handler = ClientPacketHandler(
                        reader=reader, writer=writer,
                        executor=self.executor, protocol=self.protocol,
                    )
                    if not await handler.authorize():
                        raise RuntimeError("Unauthorized")

                    on_connected.set()
                    await self._add_handler(handler)

                    await handler.start_processing()
                except asyncio.CancelledError:
                    if not writer.is_closing():
                        writer.close()
                        await writer.wait_closed()
                    raise

                if not writer.is_closing():
                    writer.close()
                    await writer.wait_closed()

                on_connected.clear()
                await self._remove_handler(handler)

                log.error(
                    "Connection to tcp://%s:%d closed. "
                    "Reconnecting after %.3f seconds...",
                    address, port, self.reconnect_timeout,
                )
            except (ConnectionError, asyncio.TimeoutError):
                log.error(
                    "Failed to establish connection to tcp://%s:%d closed. "
                    "Reconnecting after %.3f seconds.",
                    address, port, self.reconnect_timeout,
                )
            finally:
                await asyncio.sleep(self.reconnect_timeout)
            log.info(
                "Connection attempts to tcp://%s:%d has been stopped.",
                address, port,
            )

    async def connect(
        self, address: str, port: int = DEFAULT_PORT, **kwargs: Any
    ) -> None:
        event = asyncio.Event()
        self.create_task(
            self.connection_fabric(address, port, event, **kwargs),
        )
        await event.wait()


__all__ = ("TCPBrokerBase", "TCPClientBroker", "TCPServerBroker")



import asyncio
import sys
from typing import Generic, TypeVar


try:
    from typing import Self  # type: ignore
except ImportError:
    from typing_extensions import Self


T = TypeVar("T")


if sys.version_info >= (3, 9):
    from asyncio import Queue
else:
    class Queue(asyncio.Queue, Generic[T]):
        async def get(self) -> T:
            return await super().get()

        async def put(self, element: T) -> T:
            return await super().put(element)

        def get_nowait(self) -> T:
            return super().get_nowait()

        def put_nowait(self, element: T) -> None:
            return super().put_nowait(element)


__all__ = ("Queue", "Self")



from .asyncronous import AsyncExecutor
from .base import AbstractExecutor
from .null import NullExecutor
from .process_pool import ProcessPoolExecutor
from .thread_pool import ThreadPoolExecutor


__all__ = (
    "AbstractExecutor",
    "AsyncExecutor",
    "NullExecutor",
    "ProcessPoolExecutor",
    "ThreadPoolExecutor",
)



import asyncio
from functools import cached_property
from typing import Any, Callable, Set, Tuple

from patio.compat import Queue
from patio.executor.base import AbstractExecutor
from patio.registry import AsyncTaskFunctionType, T


QueueType = Queue[Tuple[Callable[..., T], Any, Any, asyncio.Future]]


class AsyncExecutor(AbstractExecutor[T]):
    """
    Executes the incoming tasks in the pool of the several concurrent tasks.
    Tasks must be asynchronous functions, or functions that return an
    awaitable object.
    This means that the whole registry must not contain functions other than
    specified kind.
    """
    __slots__ = "max_workers", "queue", "tasks"

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.queue: QueueType = Queue(maxsize=self.max_workers)
        self.tasks: Set[asyncio.Task] = set()
        self.started: bool = False

    @cached_property
    def loop(self) -> asyncio.AbstractEventLoop:
        return asyncio.get_running_loop()

    async def _executor(self) -> None:
        while True:
            func, args, kwargs, future = await self.queue.get()
            if future.done():
                continue
            try:
                future.set_result(await func(*args, **kwargs))
            except Exception as e:
                if future.done():
                    continue
                future.set_exception(e)

    async def setup(self) -> None:
        if self.started:
            return

        for _ in range(self.max_workers):
            self.tasks.add(asyncio.create_task(self._executor()))
        self.started = True

    async def submit(
        self, func: AsyncTaskFunctionType, *args: Any, **kwargs: Any
    ) -> T:
        future = self.loop.create_future()
        await self.queue.put((func, args, kwargs, future))
        return await future

    async def shutdown(self) -> None:
        cancelled = []
        for task in self.tasks:
            if task.done():
                continue
            task.cancel()
            cancelled.append(task)

        if cancelled:
            await asyncio.gather(*cancelled, return_exceptions=True)



import hashlib
import inspect
from base64 import b64encode
from collections import defaultdict
from types import MappingProxyType
from typing import (
    Any, Awaitable, Callable, DefaultDict, Dict, Generic, ItemsView, Iterator,
    KeysView, MutableMapping, Optional, Set, Tuple, TypeVar, Union, ValuesView,
    overload,
)


T = TypeVar("T")

AsyncTaskFunctionType = Callable[..., Awaitable[T]]
SyncTaskFunctionType = Callable[..., T]
TaskFunctionType = Union[AsyncTaskFunctionType, SyncTaskFunctionType]


StoreType = Union[Dict[str, TaskFunctionType], MappingProxyType]
ReverseStoreType = Union[
    DefaultDict[TaskFunctionType, Set[str]],
    MappingProxyType,
]


class Registry(MutableMapping, Generic[T]):
    """
    This is a container of functions for their subsequent execution.
    You can register a function by specific name or without it,
    in which case the function is assigned a unique name that depends on
    the source code of the function.

    This registry does not necessarily have to match on the calling and called
    sides, but for functions that you register without a name it is must be,
    and then you should not need to pass the function name but the function
    itself when you will call it.

    An instance of the registry must be transferred to the broker,
    the first broker in the process of setting up will block the registry
    to write, that is, registering new functions will be impossible.

    An optional ``project`` parameter, this is essentially like a namespace
    that will help avoid clash functions in different projects with the same
    name. It is recommended to specify it and the broker should also use this
    parameter, so it should be the same value within the same project.

    You can either manually register elements or use a
    registry instance as a decorator:

    .. code-block:: python

        from patio import Registry

        rpc = Registry(project="example")

        # Will be registered with auto generated name
        @rpc
        def mul(a, b):
            return a * b

        @rpc('div')
        def div(a, b):
            return a / b

        def pow(a, b):
            return a ** b

        def sub(a, b):
            return a - b

        # Register with auto generated name
        rpc.register(pow)

        rpc.register(sub, "sub")

    Alternatively using ``register`` method:

    .. code-block:: python

        from patio import Registry

        rpc = Registry(project="example")

        def pow(a, b):
            return a ** b

        def sub(a, b):
            return a - b

        # Register with auto generated name
        rpc.register(pow)

        rpc.register(sub, "sub")

    Finally, you can register functions explicitly, as if it were
    just a dictionary:

    .. code-block:: python

        from patio import Registry

        rpc = Registry(project="example")

        def mul(a, b):
            return a * b

        rpc['mul'] = mul

    """

    __slots__ = (
        "__project", "__strict", "__store", "__reverse_store", "__locked",
        "__auto_naming",
    )

    def __init__(
        self, project: Optional[str] = None, strict: bool = False,
        auto_naming: bool = True,
    ):
        self.__auto_naming: bool = auto_naming
        self.__project = project
        self.__strict = strict
        self.__store: StoreType = {}
        self.__reverse_store: ReverseStoreType = defaultdict(set)
        self.__locked = False

    @property
    def is_locked(self) -> bool:
        return self.__locked

    @property
    def project(self) -> Optional[str]:
        return self.__project

    @property
    def strict(self) -> bool:
        return self.__strict

    @property
    def auto_naming(self) -> bool:
        return self.__auto_naming

    def _make_function_name(self, func: TaskFunctionType) -> str:
        parts = []

        if self.project is not None:
            parts.append(self.project)

        if hasattr(func, "__module__"):
            parts.append(func.__module__)

        if hasattr(func, "__qualname__"):
            parts.append(func.__qualname__)

        if self.strict:
            sources = inspect.getsource(func)
            parts.append(
                b64encode(
                    hashlib.blake2s(sources.encode()).digest(),
                ).strip(b"=").decode(),
            )

        return ".".join(parts)

    def lock(self) -> None:
        if self.is_locked:
            return
        self.__store = MappingProxyType(self.__store)
        self.__reverse_store = MappingProxyType({
            k: frozenset(v) for k, v in self.__reverse_store.items()
        })
        self.__locked = True

    def register(
        self, func: TaskFunctionType, name: Optional[str] = None,
    ) -> str:
        if name is None:
            if not self.__auto_naming:
                raise ValueError(
                    "auto_naming is disabled, "
                    "name parameter is required",
                )
            name = self._make_function_name(func)
        self[name] = func
        return name

    @overload
    def __call__(self, name: TaskFunctionType) -> TaskFunctionType:
        ...

    @overload
    def __call__(
        self, name: Optional[str] = None,
    ) -> Callable[..., TaskFunctionType]:
        ...

    def __call__(
        self, name: Union[Optional[str], TaskFunctionType] = None,
    ) -> Union[Callable[..., TaskFunctionType], TaskFunctionType]:
        if callable(name):
            return self.__call__(None)(name)

        function_name: Optional[str] = name

        def decorator(func: TaskFunctionType) -> TaskFunctionType:
            self.register(func, function_name)
            return func
        return decorator

    def __setitem__(self, name: str, func: TaskFunctionType) -> None:
        if name in self.__store:
            raise RuntimeError(
                f"Task with name {name!r} already "
                f"registered for {self.__store[name]!r}",
            )

        if (
            isinstance(self.__store, MappingProxyType) or
            isinstance(self.__reverse_store, MappingProxyType)
        ):
            raise RuntimeError("Registry locked")

        self.__store[name] = func
        self.__reverse_store[func].add(name)

    def __delitem__(self, name: str) -> None:
        if (
            isinstance(self.__store, MappingProxyType) or
            isinstance(self.__reverse_store, MappingProxyType)
        ):
            raise RuntimeError("Registry locked")

        func = self.__store.pop(name)
        del self.__reverse_store[func]

    def __getitem__(self, name: str) -> TaskFunctionType:
        return self.__store[name]

    def __len__(self) -> int:
        return len(self.__store)

    def __iter__(self) -> Iterator[str]:
        return iter(self.__store)

    def items(self) -> ItemsView[str, TaskFunctionType]:
        return self.__store.items()

    def keys(self) -> KeysView[str]:
        return self.__store.keys()

    def values(self) -> ValuesView[TaskFunctionType]:
        return self.__store.values()

    def get_names(self, func: TaskFunctionType) -> Tuple[str, ...]:
        return tuple(self.__reverse_store[func])

    def get_name(self, func: TaskFunctionType) -> str:
        candidates = self.get_names(func)
        if not candidates:
            raise KeyError(f"{func!r} has not been registered")
        return candidates[0]

    @overload
    def resolve(self, func: str) -> Callable[..., T]:
        ...

    @overload
    def resolve(self, func: Callable[..., T]) -> Callable[..., T]:
        ...

    def resolve(
        self, func: Union[str, Callable[..., T]],
    ) -> Callable[..., Any]:
        if not isinstance(func, str):
            if not self.__auto_naming:
                raise ValueError(
                    "auto_naming is disabled, "
                    "name parameter is required",
                )
            func = self.get_name(func)
        return self[func]

    def __getstate__(self) -> Dict[str, Any]:
        return dict(
            auto_naming=self.__auto_naming,
            project=self.__project,
            strict=self.__strict,
            store=self.__store,
            locked=self.__locked,
        )

    def __setstate__(self, state: Dict[str, Any]) -> None:
        self.__auto_naming = state["auto_naming"]
        self.__project = state["project"]
        self.__strict = state["strict"]
        self.__locked = state["locked"]
        self.__store = {}
        self.__reverse_store = defaultdict(set)

        for name, func in state["store"].items():
            self[name] = func


__all__ = (
    "AsyncTaskFunctionType",
    "Registry",
    "SyncTaskFunctionType",
    "T",
    "TaskFunctionType",
)
